/*********************************************************************
*                                                                    *
*                      Software License Agreement                    *
*                                                                    *
*   The software supplied herewith by Microchip Technology           *
*   Incorporated (the "Company") for its dsPIC controller            *
*   is intended and supplied to you, the Company's customer,         *
*   for use solely and exclusively on Microchip dsPIC                *
*   products.                                                        *
*                                                                    *
*   THIS SOFTWARE IS PROVIDED IN AN "AS IS" CONDITION.  NO           *
*   WARRANTIES, WHETHER EXPRESS, IMPLIED OR STATUTORY, INCLUDING,    *
*   BUT NOT LIMITED TO, IMPLIED WARRANTIES OF MERCHANTABILITY AND    *
*   FITNESS FOR A PARTICULAR PURPOSE APPLY TO THIS SOFTWARE. THE     *
*   COMPANY SHALL NOT, IN ANY CIRCUMSTANCES, BE LIABLE FOR SPECIAL,  *
*   INCIDENTAL OR CONSEQUENTIAL DAMAGES, FOR ANY REASON WHATSOEVER.  *
*                                                                    *
*********************************************************************/

/*
** adc12.c
**
** Support for 12-bit A/D converter.
**
** The converter is set to scan, sample, and convert
** input pins AN4, AN5, AN6 automatically. These inputs
** correspond to potentiometers RP2, RP3, RP1 on the
** dsPICDEM 1.1 demo board. The sampling rate is set
** as slow as possible in order to conserve processor
** cycles for more important tasks.
**
** Each input pin is sampled twice, and an interrupt
** is generated after 6 A/D conversions.
**
** The isr routine stores an average value for each
** control in global variables RP1, RP2, RP3.
*/

#include "sound.h"
#include <p30fxxxx.h>

void delay_usec(int);

volatile int RP1, RP2, RP3;

void init_adc12()
{
    ADCON1bits.ADON = 0;    /* turn off A/D */
    ADCON1bits.ADSIDL = 0;  /* module continues when CPU is idle */
    ADCON1bits.FORM = 0;    /* format output as unsigned integer */
    ADCON1bits.SSRC = 7;    /* automatic conversions */
    ADCON1bits.ASAM = 1;    /* automatic sampling */
    
    ADCON2bits.VCFG = 0;    /* use internal voltage reference */
    ADCON2bits.CSCNA = 1;   /* scan multiple channels */
    ADCON2bits.SMPI = 6;    /* interrupt after 6 conversions */
    ADCON2bits.BUFM = 0;    /* one 16-word sample buffer */
    ADCON2bits.ALTS = 0;    /* use MUX A inputs */
    
    ADCON3bits.SAMC = 31;   /* sample for 31 Tad cycles */
    ADCON3bits.ADRC = 0;    /* derive Tad from system clock */
    ADCON3bits.ADCS = 0x3F; /* Tad equals 32 instr cycles */
    
    ADCHS = 0;              /* no static channel assignment */
    ADCSSL = 0x70;          /* select AN4, AN5, AN6 for scanning */
    
    ADPCFG = ~(0x70);       /* set AN4, AN5, AN6 as analog inputs */
    
    IFS0bits.ADIF = 0;      /* clear A/D interrupt flag */
    IPC2bits.ADIP = 4;      /* set A/D interrupt as low priority */
    IEC0bits.ADIE = 1;      /* enable A/D interrupts */
    
    ADCON1bits.ADON = 1;    /* turn on A/D */
}
    
void _ISR _ADCInterrupt()
{
    RP2 = (ADCBUF0 + ADCBUF3) / 2;
    RP3 = (ADCBUF1 + ADCBUF4) / 2;
    RP1 = (ADCBUF2 + ADCBUF5) / 2;
    IFS0bits.ADIF = 0;      /* clear A/D interrupt flag */
}

#if 0
/*
** read the potentiometers manually
*/
void read_knobs()
{
    ADCON1 = 0;
    ADPCFG = ~(0x70);       /* set AN4, AN5, AN6 as analog inputs */
    ADCSSL = 0;
    ADCON2 = 0;
    ADCON3 = 0xC;
    
    ADCHS = 0x6;            /* select RP1 input */
    ADCON1bits.ADON = 1;
    ADCON1bits.SAMP = 1;    /* start sampling */
    delay_usec(10);
    ADCON1bits.SAMP = 0;    /* start converting */
    while (!ADCON1bits.DONE); /* wait for conversion to complete */
    RP1 = ADCBUF0; 
}
#endif
