/*********************************************************************
*                                                                    *
*                      Software License Agreement                    *
*                                                                    *
*   The software supplied herewith by Microchip Technology           *
*   Incorporated (the "Company") for its dsPIC controller            *
*   is intended and supplied to you, the Company's customer,         *
*   for use solely and exclusively on Microchip dsPIC                *
*   products.                                                        *
*                                                                    *
*   THIS SOFTWARE IS PROVIDED IN AN "AS IS" CONDITION.  NO           *
*   WARRANTIES, WHETHER EXPRESS, IMPLIED OR STATUTORY, INCLUDING,    *
*   BUT NOT LIMITED TO, IMPLIED WARRANTIES OF MERCHANTABILITY AND    *
*   FITNESS FOR A PARTICULAR PURPOSE APPLY TO THIS SOFTWARE. THE     *
*   COMPANY SHALL NOT, IN ANY CIRCUMSTANCES, BE LIABLE FOR SPECIAL,  *
*   INCIDENTAL OR CONSEQUENTIAL DAMAGES, FOR ANY REASON WHATSOEVER.  *
*                                                                    *
*********************************************************************/

#include <p30fxxxx.h>
#include <dci.h>
#include "sound.h"

int interrupt_period;
int samples_per_interrupt;
int cycles_per_interrupt;
static void init_codec_slave();

/*
** init_codec()
**
** Initialization function for Si3000 codec
** on the dsPICDEM 1.1 Development Board.
** 
** Uses functions from the dsPIC Peripheral
** Library to communicate with the codec
** via the Data Converter Interface (DCI).
**
** Calls init_codec_slave() to complete
** the initialization sequence.
**
** Takes two arguments:
** 1. An integer value for the bit clock
** generator (DCICON3). See sound.h for
** suitable values.
** 2. The number of samples to process
** per interrupt (1, 2, or 4).
*/

void init_codec(unsigned int bit_clock_generator,
                unsigned int spi)
{
        /* variables to hold DCI config info */
        unsigned int DCICON1value, DCICON2value, DCICON3value;
        unsigned int TSCONvalue, RSCONvalue;
        
        samples_per_interrupt = spi;
        switch (bit_clock_generator) {
          case 2: /* 14.7 mips & 9.6Khz */ interrupt_period = 1536; break;
          case 3: /* 14.7 mips & 7.2Khz */ interrupt_period = 2048; break;
          case 4: /* 29.4 mips & 11Khz  */ interrupt_period = 2560; break;
          case 6: /* 29.4 mips & 8.2Khz */ interrupt_period = 3584; break;
        }
        //cycles_per_buffer = interrupt_period * NUM_SAMPLES;
        cycles_per_buffer = (long) interrupt_period * 256;
        /* make codec reset port pin an input for now */
        TRISFbits.TRISF6 = 1;
        
        /* ensure DCI port pins are set correctly */
        TRISGbits.TRISG12 = 1; /* CSDI configured for input  */
        TRISGbits.TRISG13 = 0; /* CSDO configured for output */
        TRISGbits.TRISG14 = 0; /* CSCK output to Si3000      */
        TRISGbits.TRISG15 = 0; /* COFS output in DCI Master mode */

        /* build the configuration words */
        DCICON1value =  DCI_FSM_MULTI &    /* multichannel frame sync */
                        DCI_DJST_OFF &     /* data begins 1 clock after Fs */
                        DCI_SDO_TRISTAT &  /* CSDO pin will be tri-stated */
                        DCI_TX_ZERO_UNF &  /* transmit zero during underflow */
                        DCI_FSD_OUP &      /* COFS pin is an output */
                        DCI_SAMP_CLK_FAL & /* sample data on falling clock edge */
                        DCI_SCKD_OUP &     /* CSCK pin is an output */
                        DCI_DIGI_LPBACK_DIS & /* loopback mode is disabled */
                        DCI_IDLE_CON &     /* DCI will continue in CPU IDLE mode */
                        DCI_DIS;           /* DCI is disabled for now */
                        
        DCICON2value =  DCI_DATA_WORD_16 & /* data word size is 16 bits */
                        DCI_FRAME_LEN_8 &  /* data frame is 8 words */
                        DCI_BUFF_LEN_2;    /* 2 words buffered per interrupt */

                                           /* calculate DCI bit clock control */
                                           /* (watch out for integer overflow) */
        DCICON3value =  bit_clock_generator;
        
        RSCONvalue   =  0x0001;            /* enable receive slot 1  */
        TSCONvalue   =  0x0001;            /* enable transmit slot 1 */
        
        IFS2bits.DCIIF = 0;                /* clear DCI interrupt flag */
        DisableIntDCI;                     /* disable DCI interrupts */

        OpenDCI(DCICON1value, DCICON2value, DCICON3value,
                TSCONvalue, RSCONvalue);
                
        init_codec_slave();                 /* initialize the codec */
        
        /* re-configure DCI for normal operation */
        DCICON2value =  DCI_DATA_WORD_16 & /* data word size is 16 bits */
                        DCI_FRAME_LEN_16; /* data frame is 16 words */

        switch (samples_per_interrupt) {
          default: samples_per_interrupt = 4;
                   /* falls through */
          case 4: DCICON2value &= DCI_BUFF_LEN_4; /* 4 words buffered per isr */
                  break;
          case 3: DCICON2value &= DCI_BUFF_LEN_3; /* 4 words buffered per isr */
                  break;
          case 2: DCICON2value &= DCI_BUFF_LEN_2; /* 4 words buffered per isr */
                  break;
          case 1: DCICON2value &= DCI_BUFF_LEN_1; /* 4 words buffered per isr */
                  break;
        }
        interrupt_period *= samples_per_interrupt;

        OpenDCI(DCICON1value, DCICON2value, DCICON3value,
                TSCONvalue, RSCONvalue);

        IPC10bits.DCIIP = 6;                /* set interrupt priority level 6 */
        EnableIntDCI;                       /* enable DCI interrupts */
        DCICON1bits.DCIEN = 1;              /* enable DCI */
} /* init_codec() */

static void init_codec_slave(void) {
  TXBUF0 = 0x7FFF;                    /* request secondary frame */
  TXBUF1 = 0x0300;                    /* initiate a write to the PLL */
  PORTFbits.RF6 = 0; 
  __asm__("nop");
  TRISFbits.TRISF6 = 0;               /* reset codec */
  __asm__("\trepeat #160\n"            /* wait for min 5uS */
          "\tnop");
  TRISFbits.TRISF6 = 1;
  
  delay_usec(150); /* DEBUG */
  
  DCICON1bits.DCIEN = 1;
  while (!DCISTATbits.TMPTY);         /* wait until TXBUF0 and TXBUF1 have been
                                         moved tot their shadow registers for
                                         transmission */
  TXBUF0 = 0x7FFF;
  TXBUF1 = 0x2300;
  while (!DCISTATbits.TMPTY);         /* wait until TXBUF0 and TXBUF1 */
  

  TXBUF0 = 0x7FFF;                    /* request secondary frame */
  TXBUF1 = 0x0413;                    /* write to PLL multiply register 4 */
  while (!DCISTATbits.TMPTY);         /* wait until TXBUF0 and TXBUF1 have been
                                         moved tot their shadow registers for
                                         transmission */
  TXBUF0 = 0x7FFF;                    /* request secondary frame */
  TXBUF1 = 0x2400;                    /* write to PLL multiply register 4 */
  while (!DCISTATbits.TMPTY);         /* wait until TXBUF0 and TXBUF1 */

  TXBUF0 = 0x7FFF;                    /* request secondary frame */
  TXBUF1 = 0x0118;                    /* TXBUF1 value that will power line drvr
                                         and speaker driver in Si3000 control 
                                         register 1 */
  while (!DCISTATbits.TMPTY);         /* wait until TXBUF0 and TXBUF1 have been
                                         moved tot their shadow registers for
                                         transmission */
  TXBUF0 = 0x7FFF;                    /* request secondary frame */
  TXBUF1 = 0x2100;                    /* write to PLL multiply register 4 */
  while (!DCISTATbits.TMPTY);         /* wait until TXBUF0 and TXBUF1 */

  TXBUF0 = 0x7FFF;                    /* request secondary frame */
  TXBUF1 = 0x0200;                    /* init control register 2 to 0 */
  while (!DCISTATbits.TMPTY);         /* wait until TXBUF0 and TXBUF1 have been
                                         moved tot their shadow registers for
                                         transmission */
  TXBUF0 = 0x7FFF;                    /* request secondary frame */
  TXBUF1 = 0x2200;                    /* write to PLL multiply register 4 */
  while (!DCISTATbits.TMPTY);         /* wait until TXBUF0 and TXBUF1 */

  TXBUF0 = 0x7FFF;                    /* request secondary frame */
  TXBUF1 = 0x0602;                    /* sets the line out active in the Si3000
                                         ADC control register 6 */
  while (!DCISTATbits.TMPTY);         /* wait until TXBUF0 and TXBUF1 have been
                                         moved tot their shadow registers for
                                         transmission */
  TXBUF0 = 0x7FFF;                    /* request secondary frame */
  TXBUF1 = 0x2600;                    /* write to PLL multiply register 4 */
  while (!DCISTATbits.TMPTY);         /* wait until TXBUF0 and TXBUF1 */

  TXBUF0 = 0x7FFF;                    /* request secondary frame */
  TXBUF1 = 0x077F;                    /* sets TX PGA gain to 12dB and activates
                                         L/R speaker in the DAC volume control
                                         register 7 */
  while (!DCISTATbits.TMPTY);         /* wait until TXBUF0 and TXBUF1 have been
                                         moved tot their shadow registers for
                                         transmission */
  TXBUF0 = 0x7FFF;                    /* request secondary frame */
  TXBUF1 = 0x2700;                    /* write to PLL multiply register 4 */
  while (!DCISTATbits.TMPTY);         /* wait until TXBUF0 and TXBUF1 */

  TXBUF0 = 0x7FFF;                    /* request secondary frame */
  TXBUF1 = 0x0583;                    /* sets LIG Gain to 10Db and mic gain to
                                         0dB, mutes Handset */
  while (!DCISTATbits.TMPTY);         /* wait until TXBUF0 and TXBUF1 have been
                                         moved tot their shadow registers for
                                         transmission */
  TXBUF0 = 0x7FFF;                    /* request secondary frame */
  TXBUF1 = 0x2700;                    /* write to PLL multiply register 4 */
  while (!DCISTATbits.TMPTY);         /* wait until TXBUF0 and TXBUF1 */


  while (DCISTAT & 0x0F00);           /* wait until all four slots have been */
  while (DCISTAT & 0x0F00);           /* transferred                         */
  while (DCISTAT & 0x0F00);
  while (DCISTAT & 0x0F00);

  if (DCISTATbits.ROV) {
    int i;

    i = RXBUF0;
    i = RXBUF1;
  }

  DCICON1bits.DCIEN = 0;

} /* init_codec_slave() */

void __attribute__((__interrupt__)) _DCIInterrupt(void) {
  static volatile unsigned int *txBuf[] = { &TXBUF0, &TXBUF1, &TXBUF2, &TXBUF3, 0 };
  static volatile unsigned int *rxBuf[] = { &RXBUF0, &RXBUF1, &RXBUF2, &RXBUF3, 0 };
  int i;
  static int d,sample_index;
  int T4CON_save;
  unsigned int isr_time;
  extern int samples[];
       
  T4CON_save = T4CON;
  T4CONbits.TON = 0;  /* pause foreground timer */
  TMR3 = 0;
  if (status.monitor || status.play) {
    if (!curr_play_buf->status & FULL) {
      /* buffer not full, play null samples */
      TXBUF0 = 0;
      TXBUF1 = 0;
      TXBUF2 = 0;
      TXBUF3 = 0;
    } else {
      int feedback_sample;
      int sample;
      unsigned int lgain = gain + 1;

      for (i = 0; i < samples_per_interrupt; i++) {
        feedback_sample =delay ? feedback_buf->sample[curr_play_buf->offset]: 0;
        feedback_sample >>= 1;
        sample =  curr_play_buf->sample[curr_play_buf->offset++];
        sample = (sample + feedback_sample) *lgain;
        *txBuf[i] = sample;
        if (sample_index < 64)
          samples[sample_index++] = sample;
      }
     
      if (curr_play_buf->offset >= NUM_SAMPLES) {
        if (curr_play_buf->status & LAST) { 
          status.play = 0;
          status.stop = 1;
        }
        curr_play_buf->status = EMPTY;
        curr_play_buf->offset = 0;
        curr_play_buf = curr_play_buf->next;
        feedback_buf = curr_play_buf;
        sample_index = 0;
        d = delay;
        for (i = 0; i < d; i++) feedback_buf = feedback_buf->prev;
      }
    }
  }
  if (status.record) {
    if (curr_rec_buf->status & REC) {
      for (i = 0; i < samples_per_interrupt; i++) {
        curr_rec_buf->sample[curr_rec_buf->offset++] = *rxBuf[i];
      }

      if (curr_rec_buf->offset >= NUM_SAMPLES) {
        if (curr_rec_buf->status & LAST) {
          status.record = 0;
          status.monitor = 0;
          status.stop = 1;
        }
        curr_rec_buf->status = FULL | NEW;
        curr_rec_buf->offset = 0;
        curr_rec_buf = curr_rec_buf->next;
        curr_rec_buf->status = REC;
      }
    }
  }
  sample_count += samples_per_interrupt;
  IFS2bits.DCIIF = 0;
  cycles_per_interrupt = TMR3 + 4; /* estimate cycles in this isr */
  T4CON = T4CON_save;  /* resume foreground timer */

} /* _DCIInterrupt */


void reset_codec()
{
  DisableIntDCI;
  PORTFbits.RF6 = 0;
  TRISFbits.TRISF6 = 0;
  delay_usec(5);
  TRISFbits.TRISF6 = 1;
}

/*
** Software delays (approximate)
*/
void delay_usec(count)
    int count;
{
        unsigned int instr = ((unsigned int)(Iusec * (count - 2)) >> 2) + 5;
        while (instr-- > 0);
}

void delay_msec(count)
    int count;
{
        unsigned int instr = ((unsigned int)(Imsec * (count - 2)) >> 2) + 5;
        while (instr-- > 0);
}

