/*********************************************************************
*                                                                    *
*                      Software License Agreement                    *
*                                                                    *
*   The software supplied herewith by Microchip Technology           *
*   Incorporated (the "Company") for its dsPIC controller            *
*   is intended and supplied to you, the Company's customer,         *
*   for use solely and exclusively on Microchip dsPIC                *
*   products.                                                        *
*                                                                    *
*   THIS SOFTWARE IS PROVIDED IN AN "AS IS" CONDITION.  NO           *
*   WARRANTIES, WHETHER EXPRESS, IMPLIED OR STATUTORY, INCLUDING,    *
*   BUT NOT LIMITED TO, IMPLIED WARRANTIES OF MERCHANTABILITY AND    *
*   FITNESS FOR A PARTICULAR PURPOSE APPLY TO THIS SOFTWARE. THE     *
*   COMPANY SHALL NOT, IN ANY CIRCUMSTANCES, BE LIABLE FOR SPECIAL,  *
*   INCIDENTAL OR CONSEQUENTIAL DAMAGES, FOR ANY REASON WHATSOEVER.  *
*                                                                    *
*********************************************************************/

/*
** frequency.c
**
** Calls an FFT function in libdsp.a
** to compute a frequency profile of
** an audio buffer. The profile is
** scaled and displayed on the lcd.
**
*/

#include "sound.h"
#include <dsp.h>
#include <p30fxxxx.h>

fractcomplex _XBSS(64) FFT_factors[32];
fractcomplex _YBSS(256) FFT_result[64];
fractional window[64];
fractional samples[64];
int profile[32];

fractcomplex *factors = (fractcomplex *) &FFT_factors;
fractcomplex *result  = (fractcomplex *) &FFT_result;

/*
** Initialize the "twiddle" factors
** and window function. Call this
** once, during startup.
*/
void init_frequency_profile() {
  factors = TwidFactorInit( 6, FFT_factors, 0);
  (void) HanningInit(64, window);
}

/*
** Clear the frequency profile
*/
void clear_frequency_profile() {
  int i;
  for (i = 0; i < 16; i++)
    profile[i] = 0;
}

/*
** Compute frequency profile on a buffer.
** In order to conserve memory, only
** the first 64 samples are considered.
** The result is a reasonable approximation
** for speech and music.
*/
void compute_frequency_profile() {
  int i;
  
 /* apply window function */
 (void) VectorWindow(64, samples, samples, window);
  
  /* copy samples into complex buffer */
  for (i=0; i<64; i++) {
    FFT_result[i].real = samples[i];
    FFT_result[i].imag = 0;
  }
	
  /* call the FFT function */
  result = FFTComplexIP( 6, FFT_result, FFT_factors, COEFFS_IN_DATA);

  /* correct profile for "bit-reversed" addressing */
  result = BitReverseComplex( 6, FFT_result);

  /*
   ** compute power spectrum
   ** Note: result[0] represents the DC component
   **       result[1..n/2] represents the spectrum
   **
   ** Actual power spectrum is the sqrt(real^2+imag^2)
   ** we save time by not finding the root, but this changes the shap of
   ** the spectrum curve.  We compensate by applying a crude scale.
   */
   for (i=1; i<32; i++) {
       static int scale[] = { 0, 8, 8, 8,
                              8, 7, 7, 6,
                              6, 6, 5, 5, 
                              5, 5 ,4, 4,
                              4, 4, 3, 3,
                              3, 3, 3, 3,
                              3, 3, 3, 3,
                              3, 3, 3, 3, };
       int s;
       int t;
       int result;
       extern int sum_of_squares(int,int);
          	
       s = FFT_result[i].real;
       t = FFT_result[i].imag;
       result = sum_of_squares(s,t);
       result <<= 8 - scale[i];
       profile[i] = (profile[i] + result)/ 2;
   }
}
