/*********************************************************************
*                                                                    *
*                      Software License Agreement                    *
*                                                                    *
*   The software supplied herewith by Microchip Technology           *
*   Incorporated (the "Company") for its dsPIC controller            *
*   is intended and supplied to you, the Company's customer,         *
*   for use solely and exclusively on Microchip dsPIC                *
*   products.                                                        *
*                                                                    *
*   THIS SOFTWARE IS PROVIDED IN AN "AS IS" CONDITION.  NO           *
*   WARRANTIES, WHETHER EXPRESS, IMPLIED OR STATUTORY, INCLUDING,    *
*   BUT NOT LIMITED TO, IMPLIED WARRANTIES OF MERCHANTABILITY AND    *
*   FITNESS FOR A PARTICULAR PURPOSE APPLY TO THIS SOFTWARE. THE     *
*   COMPANY SHALL NOT, IN ANY CIRCUMSTANCES, BE LIABLE FOR SPECIAL,  *
*   INCIDENTAL OR CONSEQUENTIAL DAMAGES, FOR ANY REASON WHATSOEVER.  *
*                                                                    *
*********************************************************************/

/*
** MASTERs 2004 Class 829
**
** Intro to Digital Sound Processing
**
** Demo Application
*/

#include <p30fxxxx.h>
#include <stdio.h>
#include <dci.h>
#include "sound.h"


/* local function prototypes */
void init_switches();
void init_status();
void init_LEDs();
void init_delay_line();
void fill_buffer_from_storage( volatile audio_bufferT *);
void fade_in( volatile audio_bufferT *);
int read_switches();
int adjust_gain();
int adjust_delay();
int adjust_samples_per();
void service_buffers();
void compute_utilization();
void start_foreground();
void stop_foreground();
void service_controls();
void update_display();
void check_for_stop();
void play(void *);
void prepare_to_stop();
void record();


/* global data structures */
volatile audio_bufferT *curr_rec_buf;
volatile audio_bufferT *curr_play_buf;
volatile audio_bufferT *feedback_buf;
storage_bufferT *storage_buf;
int	storage_offset;
volatile statusT status;
int update_freq, update_utils;

volatile unsigned int sample_count;
volatile unsigned int ticks, switch_delay;
int gain = -1, delay = -1, sample_set = -1;
int switches;

int foreground_utilization;
int background_utilization;
long cycles_per_buffer;

#define MAX 12
#define MAX_DELAY (MAX-2)
#define MAX_GAIN 8
audio_bufferT __attribute__((far)) delay_line[MAX];

int main() {
    spi_init(1);
    clear_lcd();
    delay_msec(10);
    write(0,"MCHP class 829\n",15);
    
    init_switches();
    init_LEDs();
    init_status();
    init_timer2();
    init_timer3();
    init_timer45();
    init_adc12();
    init_frequency_profile();
    
    do {
        service_buffers();
        service_controls();
        update_display();
        check_for_stop();       
    } while(1);  /* loop forever */
    return 1;
} /* main */

/*
** Service buffers in the delay line
**
** Foreground tasks are performed here.
*/
void service_buffers() {
  int buffer_serviced = 0;
  
  start_foreground();
    /*
    ** In record mode, watch for buffers
    ** marked FIRST (needs fade_in)
    ** of LAST (needs fade_out).
    */
    if (status.record) {
        audio_bufferT *rec_buf = (audio_bufferT *) curr_rec_buf;
        /* if FIRST buffer in recording */
	    if (((rec_buf->status & FIRST) == FIRST) &&
	    	(rec_buf->offset > FADE_OFFSET)) {
	    	fade_in(rec_buf);       /* avoid a pop */
	    	rec_buf->status &= ~ FIRST;
	    	buffer_serviced++;
	    }
    }
    /*
    ** In record mode, watch for buffers
    ** marked NEW (newly recorded) and run
    ** compute_frequency_profile as needed.
    */
    if (status.record) {
      audio_bufferT *rec_buf = (audio_bufferT *) curr_rec_buf;
      /* if NEW recorded buffer */
      if ((rec_buf->prev->status & NEW) == NEW) {
        compute_frequency_profile();
        rec_buf->prev->status &= ~ NEW;
	buffer_serviced++;
      }
    }
    /*
    ** In playback mode, fill only one
    ** buffer in advance. The others may
    ** be used for feedback (echo).
    */
    else if (status.play) {
        audio_bufferT *next_buf = curr_play_buf->next;
        if ((next_buf->status == EMPTY) &&
            (storage_buf != 0)) {
            fill_buffer_from_storage(next_buf);
            compute_frequency_profile();
            buffer_serviced++;
        }
    }
    stop_foreground();
    if (buffer_serviced) {
      foreground_utilization = (100*(((long) TMR5 << 16) + TMR4))
                               / cycles_per_buffer;
    }
} /* service_buffers */

/*
** Compute background utilization
** using values set by init_codec()
** and the DCI interrupt handler.
*/
void compute_utilization() {
  static int last_cycles = 0;
  
  if (last_cycles != cycles_per_interrupt) {
    background_utilization = (long) (cycles_per_interrupt * 100) / interrupt_period;
    last_cycles = cycles_per_interrupt;
  }
} /* compute_utilization */

/*
** Service the controls, including
** buttons and potentiometers.
*/
void service_controls() {   

    if (switch_delay > 0)       /* is delay complete? */
        return;                 /* exit if not        */

  /* gain and delay can be adjusted anytime */
    if (adjust_gain() | adjust_delay())
      update_control_display1();

  /* sample_set can be adjusted only when idle */
    if (adjust_sample_set())
      update_control_display2();
    
    switches = read_switches();
    if (switches == 0)
        return;
        
    switch_delay = 25;          /* set delay = 250 ms */
    
    /* any button stops recording */
    if (status.record == 1) {
        DisableIntDCI;
        init_status();
        cycles_per_interrupt = 0;
        update_control_display2();
        clear_frequency_profile();
        return;
    }     
    if ((switches == 1) && (status.play == 0)) {
        play((void *) &track1);
    }
    else if ((switches == 2) && (status.play == 0)) {
        play((void *) &track2);
    }      
    else if ((switches == 4) && (status.play == 0)) {
        play((void *) &track3);
    }     
    else if (switches == 8) {
        record();
    }
} /* service_controls */

void check_for_stop() {
    if (status.stop == 1) {
        DisableIntDCI;
        status.stop = 0;
        status.play = 0;
        storage_buf = 0;
        cycles_per_interrupt = 0;
        update_control_display2();
        clear_frequency_profile();
    }
} /* check_for_stop */

/*
** Play a track from program memory.
**
** Memory tracks are recorded at 8 Khz,
** so we initialize the codec at LOW_RATE.
*/
void play(void *sound)
{
    init_delay_line();
    storage_buf = (storage_bufferT *) sound;
    storage_offset = STORAGE_HEADER_SIZE;
    fill_buffer_from_storage(curr_play_buf);
    fade_in(curr_play_buf);
    status.play = 1;
    reset_utilization();
    init_codec(LOW_RATE,sample_set);
} /* play */

/*
** Turn on record function (pass-through).
**
** Pass through sound can be processed at
** any sample rate supported by the system.
** Initialize the codec at HIGH_RATE or
** LOW_RATE.
*/
void record() {
    init_delay_line();
    curr_rec_buf->status |= REC;
    status.record = 1;
    status.monitor = 1;
    init_codec(HIGH_RATE,sample_set);
    reset_utilization();
    /* init_codec(LOW_RATE,sample_set); */
} /* record */

/*
** Fill a buffer from program memory.
**
** When a track is complete, pad the
** last buffer to make it full, and
** then fill a "tail" of delay+1 buffers.
** This allocates space/time for the echo.
*/
void fill_buffer_from_storage(volatile audio_bufferT * buf) {
  char *source_ptr = (char *) storage_buf + storage_offset;
  int *dest_ptr   = buf->sample;
  int n_samples = storage_buf->length - storage_offset;
  int pad = 0;
  static int tail = 0;

  if (tail) {
    pad = NUM_SAMPLES;
    tail--;
    n_samples = 0;
    if (tail == 0) {
      buf->status |= LAST;
      storage_buf = 0;
    }
  } else if (n_samples < NUM_SAMPLES) {
    pad = NUM_SAMPLES - n_samples;
    storage_offset = STORAGE_HEADER_SIZE;
    tail = delay+1;
  } else {
    n_samples = NUM_SAMPLES;
    storage_offset += n_samples;
  }   

  while (n_samples-- > 0)   /* copy samples */
    *dest_ptr++ = (int) *source_ptr++;

  while (pad-- > 0)       /* pad with silence */
    *dest_ptr++ = 0;

  buf->status |= FULL;
  buf->offset = 0;
} /* fill_buffer_from_storage */


/*
** Fade in a buffer to avoid snaps.
*/
void fade_in(buf)
	volatile audio_bufferT *buf;
{
	int i;
	const int steps = FADE_OFFSET/2;
	for (i=0; i<steps; i++) {
    	int s = buf->sample[i]/steps;
	    buf->sample[i] = (s * i);
	}
} /* fade_in */

/*
** Initialize the delay line.
*/
void init_delay_line()
{
    int i,j;

    curr_rec_buf = &delay_line[0];
    curr_play_buf = &delay_line[0];
    feedback_buf = &delay_line[0];

    for (i = 0; i < MAX; i++)
    {
        delay_line[i].status = EMPTY;
        delay_line[i].offset = 0;

        /* set the forward links */
        if (i < (MAX -1))
            delay_line[i].next = &delay_line[i+1];
        else
            delay_line[i].next = &delay_line[0];

        /* set the reverse links */
        if (i > 0)
            delay_line[i].prev = &delay_line[i-1];
        else
            delay_line[i].prev = &delay_line[MAX-1];

        /* clear the sample area */
        for (j = 0; j < NUM_SAMPLES; j++)
            delay_line[i].sample[j] = 0;
    }
    /* mark first buffer to suppress pop */
    delay_line[0].status |= FIRST;
} /* init_delay_line */

/*
** Read the push buttons.
*/
int read_switches() {
    return (~(PORTA & 0xF000)) >> 12;
} /* read_switches */

/*
** Read potentiometer #1 and
** adjust gain setting.
*/
int adjust_gain() {
    int rp = RP1;
    rp = (rp > 0x800) ? (rp / 0x80) - 16 : 0;
    if (rp > MAX_GAIN)
        rp = MAX_GAIN;
    if (rp != gain) {
      gain = rp;
      return 1;
    }
    return 0;
} /* adjust_gain */

/*
** Read potentiometer #2 and
** adjust delay setting.
**
** Delay is the number of buffers
** that are "looked back" in order
** to add an echo.
*/
int adjust_delay() {
    int rp = (RP2 > 0x800) ? (RP2 / 0x80) - 16 : 0;
    if (rp > MAX_DELAY)
        rp = MAX_DELAY;
    if (rp != delay) {
      delay = rp;
      return 1;
    }
    return 0;
} /* adjust_delay */

/*
** Read potentiometer #3 and
** adjust sample_set setting.
**
** sample_set is passed to init_codec()
** and determines how many samples
** are read/written for each DCI interrupt.
*/
int adjust_sample_set() {
    int rp = (RP3 > 0x800) ? (RP3 / 0x200) - 3 : 1;
    if (rp > 2)
        rp = 4;
    if (rp != sample_set) {
      sample_set = rp;
      return 1;
    }
    return 0;
} /* adjust_sample_set */

/*
** Start timing a foreground task.
*/
void start_foreground() {
    TMR4 = 0;
    TMR5 = 0;
    T4CONbits.TON = 1;
} /* start_foreground */

/*
** Stop timing a foreground task.
*/
void stop_foreground() {
    T4CONbits.TON = 0;
}

/*
** Initialize the system status.
*/
void init_status() {
    status.record = 0;
    status.monitor = 0;
    status.play = 0;
    storage_buf = 0;
} /* init_status */

/*
** Initialize the buttons.
*/
void init_switches() {
    TRISAbits.TRISA12 = 1;
    TRISAbits.TRISA13 = 1;
    TRISAbits.TRISA14 = 1;
    TRISAbits.TRISA15 = 1;
} /* init_switches */

/*
** Initialize the LEDs.
*/
void init_LEDs() {
    PORTDbits.RD0 = 1;
    PORTDbits.RD1 = 1;
    PORTDbits.RD2 = 1;
    PORTDbits.RD3 = 1;
    
    TRISDbits.TRISD0 = 0;
    TRISDbits.TRISD1 = 0;
    TRISDbits.TRISD2 = 0;
    TRISDbits.TRISD3 = 0;
} /* init_LEDs */
	
